# Copyright 2023 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Component that chain components in a sequential way, removing concurrency."""

from typing import Sequence

from concordia.typing import component


class Sequential(component.Component):
  """Chains components, removing concurrency."""

  def __init__(self, name: str, components: Sequence[component.Component]):
    self._components = components
    self._name = name

  def update(self) -> None:
    for comp in self._components:
      comp.update()

  def state(self) -> str:
    return '\n'.join(
        [comp.name() + ': ' + comp.state() for comp in self._components]
    )

  def partial_state(self, player_name: str) -> str | None:
    return '\n'.join(
        [comp.partial_state(player_name) for comp in self._components]
    )

  def observe(self, observation: str):
    for comp in self._components:
      comp.observe(observation)

  def update_before_event(self, cause_statement: str) -> None:
    for comp in self._components:
      comp.update_before_event(cause_statement)

  def update_after_event(self, event_statement: str) -> None:
    for comp in self._components:
      comp.update_after_event(event_statement)

  def terminate_episode(self) -> bool:
    for comp in self._components:
      if comp.terminate_episode():
        return True
    return False

  def name(self) -> str:
    return self._name

  def get_last_log(
      self,
  ):
    """Returns a dictionary with latest log of activity."""
    output = {}
    for comp in self._components:
      output[comp.name()] = comp.get_last_log()

    return output
